@@ -16,25 +16,28 @@
  * along with sxiv.  If not, see <http://www.gnu.org/licenses/>.
  */
 
-#define _POSIX_C_SOURCE 200112L
-#define _IMAGE_CONFIG
-
+#include <errno.h>
 #include <stdlib.h>
 #include <string.h>
 #include <sys/types.h>
 #include <unistd.h>
 
-#if HAVE_GIFLIB
-#include <gif_lib.h>
-enum { MIN_GIF_DELAY = 25 };
-#endif
-
-#include "exif.h"
 #include "image.h"
 #include "options.h"
 #include "util.h"
+
+#define _IMAGE_CONFIG
 #include "config.h"
 
+#if HAVE_LIBEXIF
+#include <libexif/exif-data.h>
+#endif
+
+#if HAVE_GIFLIB
+#include <gif_lib.h>
+enum { DEF_GIF_DELAY = 75 };
+#endif
+
 float zoom_min;
 float zoom_max;
 
@@ -43,30 +46,11 @@ static int zoomdiff(float z1, float z2)
 	return (int) (z1 * 1000.0 - z2 * 1000.0);
 }
 
-static void img_apply_gamma(img_t *img)
-{
-	if (img == NULL || img->im == NULL || img->cmod == NULL)
-		return;
-	
-	if (img->gamma == 0) {
-		imlib_context_set_color_modifier(NULL);
-	} else {
-		double range = img->gamma <= 0 ? 1.0 : GAMMA_MAX - 1.0;
-
-		imlib_context_set_color_modifier(img->cmod);
-		imlib_reset_color_modifier();
-		imlib_modify_color_modifier_gamma(1.0 + img->gamma * (range / GAMMA_RANGE));
-	}
-}
-
 void img_init(img_t *img, win_t *win)
 {
 	zoom_min = zoom_levels[0] / 100.0;
 	zoom_max = zoom_levels[ARRLEN(zoom_levels) - 1] / 100.0;
 
-	if (img == NULL || win == NULL)
-		return;
-
 	imlib_context_set_display(win->env.dpy);
 	imlib_context_set_visual(win->env.vis);
 	imlib_context_set_colormap(win->env.cmap);
@@ -82,19 +66,33 @@ void img_init(img_t *img, win_t *win)
 	img->aa = ANTI_ALIAS;
 	img->alpha = ALPHA_LAYER;
 	img->multi.cap = img->multi.cnt = 0;
-	img->multi.animate = false;
-	img->multi.length = img->multi.repeat = 0;
+	img->multi.animate = options->animate;
+	img->multi.length = 0;
 
 	img->cmod = imlib_create_color_modifier();
+	imlib_context_set_color_modifier(img->cmod);
 	img->gamma = MIN(MAX(options->gamma, -GAMMA_RANGE), GAMMA_RANGE);
-	
+
 	img->ss.on = options->slideshow > 0;
 	img->ss.delay = options->slideshow > 0 ? options->slideshow : SLIDESHOW_DELAY;
 }
 
+#if HAVE_LIBEXIF
 void exif_auto_orientate(const fileinfo_t *file)
 {
-	switch (exif_orientation(file)) {
+	ExifData *ed;
+	ExifEntry *entry;
+	int byte_order, orientation = 0;
+
+	if ((ed = exif_data_new_from_file(file->path)) == NULL)
+		return;
+	byte_order = exif_data_get_byte_order(ed);
+	entry = exif_content_get_entry(ed->ifd[EXIF_IFD_0], EXIF_TAG_ORIENTATION);
+	if (entry != NULL)
+		orientation = exif_get_short(entry->data, byte_order);
+	exif_data_unref(ed);
+
+	switch (orientation) {
 		case 5:
 			imlib_image_orientate(1);
 		case 2:
@@ -116,6 +114,7 @@ void exif_auto_orientate(const fileinfo_t *file)
 			break;
 	}
 }
+#endif
 
 #if HAVE_GIFLIB
 bool img_load_gif(img_t *img, const fileinfo_t *file)
@@ -140,10 +139,10 @@ bool img_load_gif(img_t *img, const fileinfo_t *file)
 	if (img->multi.cap == 0) {
 		img->multi.cap = 8;
 		img->multi.frames = (img_frame_t*)
-		                    s_malloc(sizeof(img_frame_t) * img->multi.cap);
+		                    emalloc(sizeof(img_frame_t) * img->multi.cap);
 	}
 	img->multi.cnt = img->multi.sel = 0;
-	img->multi.length = img->multi.repeat = 0;
+	img->multi.length = 0;
 
 #if defined(GIFLIB_MAJOR) && GIFLIB_MAJOR >= 5
 	gif = DGifOpenFileName(file->path, NULL);
@@ -151,7 +150,7 @@ bool img_load_gif(img_t *img, const fileinfo_t *file)
 	gif = DGifOpenFileName(file->path);
 #endif
 	if (gif == NULL) {
-		warn("could not open gif file: %s", file->name);
+		error(0, 0, "%s: Error opening gif image", file->name);
 		return false;
 	}
 	bg = gif->SBackGroundColor;
@@ -177,16 +176,7 @@ bool img_load_gif(img_t *img, const fileinfo_t *file)
 						transp = -1;
 
 					delay = 10 * ((unsigned int) ext[3] << 8 | (unsigned int) ext[2]);
-					if (delay)
-						delay = MAX(delay, MIN_GIF_DELAY);
-
 					disposal = (unsigned int) ext[1] >> 2 & 0x7;
-				} else if (ext_code == APPLICATION_EXT_FUNC_CODE) {
-					if (ext[0] == 11 && memcmp(ext+1, "NETSCAPE2.0", 11) == 0) {
-						DGifGetExtensionNext(gif, &ext);
-						if (ext && ext[0] == 3 && ext[1] == 1)
-							img->multi.repeat = ((int) ext[3] << 8 | (int) ext[2]) - 1;
-					}
 				}
 				ext = NULL;
 				DGifGetExtensionNext(gif, &ext);
@@ -201,9 +191,9 @@ bool img_load_gif(img_t *img, const fileinfo_t *file)
 			w = gif->Image.Width;
 			h = gif->Image.Height;
 
-			rows = (GifRowType*) s_malloc(h * sizeof(GifRowType));
+			rows = (GifRowType*) emalloc(h * sizeof(GifRowType));
 			for (i = 0; i < h; i++)
-				rows[i] = (GifRowType) s_malloc(w * sizeof(GifPixelType));
+				rows[i] = (GifRowType) emalloc(w * sizeof(GifPixelType));
 			if (gif->Image.Interlace) {
 				for (i = 0; i < 4; i++) {
 					for (j = intoffset[i]; j < h; j += intjump[i])
@@ -214,7 +204,7 @@ bool img_load_gif(img_t *img, const fileinfo_t *file)
 					DGifGetLine(gif, rows[i], w);
 			}
 
-			ptr = data = (DATA32*) s_malloc(sizeof(DATA32) * sw * sh);
+			ptr = data = (DATA32*) emalloc(sizeof(DATA32) * sw * sh);
 			cmap = gif->Image.ColorMap ? gif->Image.ColorMap : gif->SColorMap;
 			r = cmap->Colors[bg].Red;
 			g = cmap->Colors[bg].Green;
@@ -268,31 +258,33 @@ bool img_load_gif(img_t *img, const fileinfo_t *file)
 			if (img->multi.cnt == img->multi.cap) {
 				img->multi.cap *= 2;
 				img->multi.frames = (img_frame_t*)
-				                    s_realloc(img->multi.frames,
-				                              img->multi.cap * sizeof(img_frame_t));
+				                    erealloc(img->multi.frames,
+				                             img->multi.cap * sizeof(img_frame_t));
 			}
 			img->multi.frames[img->multi.cnt].im = im;
-			img->multi.frames[img->multi.cnt].delay = delay ? delay : GIF_DELAY;
+			img->multi.frames[img->multi.cnt].delay = delay > 0 ? delay : DEF_GIF_DELAY;
 			img->multi.length += img->multi.frames[img->multi.cnt].delay;
 			img->multi.cnt++;
 		}
 	} while (rec != TERMINATE_RECORD_TYPE);
 
+#if defined(GIFLIB_MAJOR) && GIFLIB_MAJOR >= 5 && GIFLIB_MINOR >= 1
+	DGifCloseFile(gif, NULL);
+#else
 	DGifCloseFile(gif);
+#endif
 
-	if (err && !file->loaded)
-		warn("corrupted gif file: %s", file->name);
+	if (err && (file->flags & FF_WARN))
+		error(0, 0, "%s: Corrupted gif file", file->name);
 
 	if (img->multi.cnt > 1) {
 		imlib_context_set_image(img->im);
 		imlib_free_image();
 		img->im = img->multi.frames[0].im;
-		img->multi.animate = GIF_AUTOPLAY;
 	} else if (img->multi.cnt == 1) {
 		imlib_context_set_image(img->multi.frames[0].im);
 		imlib_free_image();
 		img->multi.cnt = 0;
-		img->multi.animate = false;
 	}
 
 	imlib_context_set_image(img->im);
@@ -305,32 +297,27 @@ bool img_load(img_t *img, const fileinfo_t *file)
 {
 	const char *fmt;
 
-	if (img == NULL || file == NULL || file->name == NULL || file->path == NULL)
-		return false;
-
 	if (access(file->path, R_OK) < 0 ||
 	    (img->im = imlib_load_image(file->path)) == NULL)
 	{
-		warn("could not open image: %s", file->name);
+		if (file->flags & FF_WARN)
+			error(0, 0, "%s: Error opening image", file->name);
 		return false;
 	}
 
 	imlib_context_set_image(img->im);
 	imlib_image_set_changes_on_disk();
 
-	if ((fmt = imlib_image_format()) == NULL) {
-		warn("could not open image: %s", file->name);
-		return false;
-	}
-	if (STREQ(fmt, "jpeg"))
-		exif_auto_orientate(file);
-#if HAVE_GIFLIB
-	if (STREQ(fmt, "gif"))
-		img_load_gif(img, file);
+#if HAVE_LIBEXIF
+	exif_auto_orientate(file);
 #endif
 
-	img_apply_gamma(img);
-
+	if ((fmt = imlib_image_format()) != NULL) {
+#if HAVE_GIFLIB
+		if (STREQ(fmt, "gif"))
+			img_load_gif(img, file);
+#endif
+	}
 	img->w = imlib_image_get_width();
 	img->h = imlib_image_get_height();
 	img->checkpan = true;
@@ -339,13 +326,10 @@ bool img_load(img_t *img, const fileinfo_t *file)
 	return true;
 }
 
-void img_close(img_t *img, bool decache)
+CLEANUP void img_close(img_t *img, bool decache)
 {
 	int i;
 
-	if (img == NULL)
-		return;
-
 	if (img->multi.cnt > 0) {
 		for (i = 0; i < img->multi.cnt; i++) {
 			imlib_context_set_image(img->multi.frames[i].im);
@@ -361,9 +345,6 @@ void img_close(img_t *img, bool decache)
 			imlib_free_image();
 		img->im = NULL;
 	}
-
-	if (img->cmod)
-		imlib_context_set_color_modifier(NULL);
 }
 
 void img_check_pan(img_t *img, bool moved)
@@ -372,9 +353,6 @@ void img_check_pan(img_t *img, bool moved)
 	int ox, oy;
 	float w, h;
 
-	if (img == NULL || img->im == NULL || img->win == NULL)
-		return;
-
 	win = img->win;
 	w = img->w * img->zoom;
 	h = img->h * img->zoom;
@@ -402,8 +380,6 @@ bool img_fit(img_t *img)
 {
 	float z, zmax, zw, zh;
 
-	if (img == NULL || img->im == NULL || img->win == NULL)
-		return false;
 	if (img->scalemode == SCALE_ZOOM)
 		return false;
 
@@ -443,9 +419,6 @@ void img_render(img_t *img)
 	Imlib_Image bg;
 	unsigned long c;
 
-	if (img == NULL || img->im == NULL || img->win == NULL)
-		return;
-
 	win = img->win;
 	img_fit(img);
 
@@ -462,7 +435,7 @@ void img_render(img_t *img)
 	 *   - full image drawn on part of window
 	 */
 	if (img->x <= 0) {
-		sx = -img->x / img->zoom;
+		sx = -img->x / img->zoom + 0.5;
 		sw = win->w / img->zoom;
 		dx = 0;
 		dw = win->w;
@@ -473,7 +446,7 @@ void img_render(img_t *img)
 		dw = img->w * img->zoom;
 	}
 	if (img->y <= 0) {
-		sy = -img->y / img->zoom;
+		sy = -img->y / img->zoom + 0.5;
 		sh = win->h / img->zoom;
 		dy = 0;
 		dh = win->h;
@@ -488,11 +461,11 @@ void img_render(img_t *img)
 
 	imlib_context_set_image(img->im);
 	imlib_context_set_anti_alias(img->aa);
-	imlib_context_set_drawable(win->pm);
+	imlib_context_set_drawable(win->buf.pm);
 
 	if (imlib_image_has_alpha()) {
 		if ((bg = imlib_create_image(dw, dh)) == NULL)
-			die("could not allocate memory");
+			error(EXIT_FAILURE, ENOMEM, NULL);
 		imlib_context_set_image(bg);
 		imlib_image_set_has_alpha(0);
 
@@ -519,10 +492,8 @@ void img_render(img_t *img)
 		imlib_blend_image_onto_image(img->im, 0, sx, sy, sw, sh, 0, 0, dw, dh);
 		imlib_context_set_color_modifier(NULL);
 		imlib_render_image_on_drawable(dx, dy);
-
 		imlib_free_image();
-		if (img->gamma != 0)
-			imlib_context_set_color_modifier(img->cmod);
+		imlib_context_set_color_modifier(img->cmod);
 	} else {
 		imlib_render_image_part_on_drawable_at_size(sx, sy, sw, sh, dx, dy, dw, dh);
 	}
@@ -533,9 +504,6 @@ bool img_fit_win(img_t *img, scalemode_t sm)
 {
 	float oz;
 
-	if (img == NULL || img->im == NULL || img->win == NULL)
-		return false;
-
 	oz = img->zoom;
 	img->scalemode = sm;
 
@@ -551,9 +519,6 @@ bool img_fit_win(img_t *img, scalemode_t sm)
 
 bool img_zoom(img_t *img, float z)
 {
-	if (img == NULL || img->im == NULL || img->win == NULL)
-		return false;
-
 	z = MAX(z, zoom_min);
 	z = MIN(z, zoom_max);
 
@@ -576,9 +541,6 @@ bool img_zoom_in(img_t *img)
 	int i;
 	float z;
 
-	if (img == NULL || img->im == NULL)
-		return false;
-
 	for (i = 1; i < ARRLEN(zoom_levels); i++) {
 		z = zoom_levels[i] / 100.0;
 		if (zoomdiff(z, img->zoom) > 0)
@@ -592,9 +554,6 @@ bool img_zoom_out(img_t *img)
 	int i;
 	float z;
 
-	if (img == NULL || img->im == NULL)
-		return false;
-
 	for (i = ARRLEN(zoom_levels) - 2; i >= 0; i--) {
 		z = zoom_levels[i] / 100.0;
 		if (zoomdiff(z, img->zoom) < 0)
@@ -607,9 +566,6 @@ bool img_move(img_t *img, float dx, float dy)
 {
 	float ox, oy;
 
-	if (img == NULL || img->im == NULL)
-		return false;
-
 	ox = img->x;
 	oy = img->y;
 
@@ -634,9 +590,6 @@ bool img_pan(img_t *img, direction_t dir, int d)
 	 */
 	float x, y;
 
-	if (img == NULL || img->im == NULL || img->win == NULL)
-		return false;
-
 	if (d > 0) {
 		x = y = MAX(1, (float) d * img->zoom);
 	} else {
@@ -661,26 +614,17 @@ bool img_pan_edge(img_t *img, direction_t dir)
 {
 	int ox, oy;
 
-	if (img == NULL || img->im == NULL || img->win == NULL)
-		return false;
-
 	ox = img->x;
 	oy = img->y;
 
-	switch (dir) {
-		case DIR_LEFT:
-			img->x = 0;
-			break;
-		case DIR_RIGHT:
-			img->x = img->win->w - img->w * img->zoom;
-			break;
-		case DIR_UP:
-			img->y = 0;
-			break;
-		case DIR_DOWN:
-			img->y = img->win->h - img->h * img->zoom;
-			break;
-	}
+	if (dir & DIR_LEFT)
+		img->x = 0;
+	if (dir & DIR_RIGHT)
+		img->x = img->win->w - img->w * img->zoom;
+	if (dir & DIR_UP)
+		img->y = 0;
+	if (dir & DIR_DOWN)
+		img->y = img->win->h - img->h * img->zoom;
 
 	img_check_pan(img, true);
 
@@ -696,9 +640,6 @@ void img_rotate(img_t *img, degree_t d)
 {
 	int i, ox, oy, tmp;
 
-	if (img == NULL || img->im == NULL || img->win == NULL)
-		return;
-
 	imlib_context_set_image(img->im);
 	imlib_image_orientate(d);
 
@@ -734,7 +675,7 @@ void img_flip(img_t *img, flipdir_t d)
 
 	d = (d & (FLIP_HORIZONTAL | FLIP_VERTICAL)) - 1;
 
-	if (img == NULL || img->im == NULL || d < 0 || d >= ARRLEN(imlib_flip_op))
+	if (d < 0 || d >= ARRLEN(imlib_flip_op))
 		return;
 
 	imlib_context_set_image(img->im);
@@ -751,9 +692,6 @@ void img_flip(img_t *img, flipdir_t d)
 
 void img_toggle_antialias(img_t *img)
 {
-	if (img == NULL || img->im == NULL)
-		return;
-
 	img->aa = !img->aa;
 	imlib_context_set_image(img->im);
 	imlib_context_set_anti_alias(img->aa);
@@ -767,20 +705,20 @@ bool img_change_gamma(img_t *img, int d)
 	 * d > 0: increase gamma
 	 */
 	int gamma;
-
-	if (img == NULL || img->im == NULL)
-		return false;
+	double range;
 
 	if (d == 0)
 		gamma = 0;
-	else if (d < 0)
-		gamma = MAX(-GAMMA_RANGE, img->gamma - 1);
 	else
-		gamma = MIN(+GAMMA_RANGE, img->gamma + 1);
+		gamma = MIN(MAX(img->gamma + d, -GAMMA_RANGE), GAMMA_RANGE);
 
 	if (img->gamma != gamma) {
+		imlib_reset_color_modifier();
+		if (gamma != 0) {
+			range = gamma <= 0 ? 1.0 : GAMMA_MAX - 1.0;
+			imlib_modify_color_modifier_gamma(1.0 + gamma * (range / GAMMA_RANGE));
+		}
 		img->gamma = gamma;
-		img_apply_gamma(img);
 		img->dirty = true;
 		return true;
 	} else {
@@ -790,8 +728,6 @@ bool img_change_gamma(img_t *img, int d)
 
 bool img_frame_goto(img_t *img, int n)
 {
-	if (img == NULL || img->im == NULL)
-		return false;
 	if (n < 0 || n >= img->multi.cnt || n == img->multi.sel)
 		return false;
 
@@ -809,7 +745,7 @@ bool img_frame_goto(img_t *img, int n)
 
 bool img_frame_navigate(img_t *img, int d)
 {
-	if (img == NULL|| img->im == NULL || img->multi.cnt == 0 || d == 0)
+	if (img->multi.cnt == 0 || d == 0)
 		return false;
 
 	d += img->multi.sel;
@@ -821,27 +757,16 @@ bool img_frame_navigate(img_t *img, int d)
 	return img_frame_goto(img, d);
 }
 
-bool img_frame_animate(img_t *img, bool restart)
+bool img_frame_animate(img_t *img)
 {
-	if (img == NULL || img->im == NULL || img->multi.cnt == 0)
+	if (img->multi.cnt == 0)
 		return false;
 
-	if (img->multi.sel + 1 >= img->multi.cnt) {
-		if (restart || GIF_LOOP == 1) {
-			img_frame_goto(img, 0);
-		} else if (GIF_LOOP == -1 && img->multi.repeat != 0) {
-			if (img->multi.repeat > 0)
-				img->multi.repeat--;
-			img_frame_goto(img, 0);
-		} else {
-			img->multi.animate = false;
-			return false;
-		}
-	} else if (!restart) {
+	if (img->multi.sel + 1 >= img->multi.cnt)
+		img_frame_goto(img, 0);
+	else
 		img_frame_goto(img, img->multi.sel + 1);
-	}
-	img->multi.animate = true;
 	img->dirty = true;
-
 	return true;
 }
+
