@@ -36,7 +36,6 @@ typedef struct {
 	int sel;
 	bool animate;
 	int length;
-	int repeat;
 } multi_img_t;
 
 typedef struct {
@@ -51,7 +50,6 @@ typedef struct {
 	scalemode_t scalemode;
 	float zoom;
 
-	bool re;
 	bool checkpan;
 	bool dirty;
 	bool aa;
@@ -71,7 +69,7 @@ typedef struct {
 void img_init(img_t*, win_t*);
 
 bool img_load(img_t*, const fileinfo_t*);
-void img_close(img_t*, bool);
+CLEANUP void img_close(img_t*, bool);
 
 void img_render(img_t*);
 
@@ -93,6 +91,6 @@ void img_toggle_antialias(img_t*);
 bool img_change_gamma(img_t*, int);
 
 bool img_frame_navigate(img_t*, int);
-bool img_frame_animate(img_t*, bool);
+bool img_frame_animate(img_t*);
 
 #endif /* IMAGE_H */
