@@ -16,9 +16,6 @@
  * along with sxiv.  If not, see <http://www.gnu.org/licenses/>.
  */
 
-#define _POSIX_C_SOURCE 200112L
-#define _MAPPINGS_CONFIG
-
 #include <stdlib.h>
 #include <stdio.h>
 #include <stdarg.h>
@@ -41,18 +38,10 @@
 #include "thumbs.h"
 #include "util.h"
 #include "window.h"
+
+#define _MAPPINGS_CONFIG
 #include "config.h"
 
-enum {
-	FILENAME_CNT = 1024,
-	TITLE_LEN    = 256
-};
-
-typedef struct {
-	const char *name;
-	char *cmd;
-} exec_t;
-
 typedef struct {
 	struct timeval when;
 	bool active;
@@ -81,15 +70,20 @@ bool extprefix;
 
 bool resized = false;
 
+typedef struct {
+	int err;
+	char *cmd;
+} extcmd_t;
+
 struct {
-  char *cmd;
+  extcmd_t f;
   int fd;
   unsigned int i, lastsep;
   bool open;
 } info;
 
 struct {
-	char *cmd;
+	extcmd_t f;
 	bool warned;
 } keyhandler;
 
@@ -103,60 +97,41 @@ timeout_t timeouts[] = {
 
 void cleanup(void)
 {
-	static bool in = false;
-
-	if (!in) {
-		in = true;
-		img_close(&img, false);
-		tns_free(&tns);
-		win_close(&win);
-	}
+	img_close(&img, false);
+	tns_free(&tns);
+	win_close(&win);
 }
 
 void check_add_file(char *filename, bool given)
 {
+	char *path;
 	const char *bn;
 
-	if (filename == NULL || *filename == '\0')
+	if (*filename == '\0')
 		return;
 
-	if (access(filename, R_OK) < 0) {
+	if (access(filename, R_OK) < 0 ||
+	    (path = realpath(filename, NULL)) == NULL)
+	{
 		if (given)
-			warn("could not open file: %s", filename);
+			error(0, errno, "%s", filename);
 		return;
 	}
 
 	if (fileidx == filecnt) {
 		filecnt *= 2;
-		files = (fileinfo_t*) s_realloc(files, filecnt * sizeof(fileinfo_t));
+		files = erealloc(files, filecnt * sizeof(*files));
+		memset(&files[filecnt/2], 0, filecnt/2 * sizeof(*files));
 	}
 
-#if defined _BSD_SOURCE || defined _XOPEN_SOURCE && \
-    ((_XOPEN_SOURCE - 0) >= 500 || defined _XOPEN_SOURCE_EXTENDED)
-
-	if ((files[fileidx].path = realpath(filename, NULL)) == NULL) {
-		warn("could not get real path of file: %s\n", filename);
-		return;
-	}
-#else
-	if (*filename != '/') {
-		if ((files[fileidx].path = absolute_path(filename)) == NULL) {
-			warn("could not get absolute path of file: %s\n", filename);
-			return;
-		}
-	} else {
-		files[fileidx].path = NULL;
-	}
-#endif
-
-	files[fileidx].warn = given;
-	files[fileidx].name = s_strdup(filename);
-	if (files[fileidx].path == NULL)
-		files[fileidx].path = files[fileidx].name;
+	files[fileidx].name = estrdup(filename);
+	files[fileidx].path = path;
 	if ((bn = strrchr(files[fileidx].name , '/')) != NULL && bn[1] != '\0')
 		files[fileidx].base = ++bn;
 	else
 		files[fileidx].base = files[fileidx].name;
+	if (given)
+		files[fileidx].flags |= FF_WARN;
 	fileidx++;
 }
 
@@ -168,10 +143,9 @@ void remove_file(int n, bool manual)
 	if (filecnt == 1) {
 		if (!manual)
 			fprintf(stderr, "sxiv: no more files to display, aborting\n");
-		cleanup();
 		exit(manual ? EXIT_SUCCESS : EXIT_FAILURE);
 	}
-	if (files[n].marked)
+	if (files[n].flags & FF_MARK)
 		markcnt--;
 
 	if (files[n].path != files[n].name)
@@ -251,7 +225,7 @@ void open_info(void)
 	static pid_t pid;
 	int pfd[2];
 
-	if (info.cmd == NULL || info.open || win.bar.h == 0)
+	if (info.f.err != 0 || info.open || win.bar.h == 0)
 		return;
 	if (info.fd != -1) {
 		close(info.fd);
@@ -262,19 +236,20 @@ void open_info(void)
 
 	if (pipe(pfd) < 0)
 		return;
-	pid = fork();
-	if (pid > 0) {
-		close(pfd[1]);
+	if ((pid = fork()) == 0) {
+		close(pfd[0]);
+		dup2(pfd[1], 1);
+		execl(info.f.cmd, info.f.cmd, files[fileidx].name, NULL);
+		error(EXIT_FAILURE, errno, "exec: %s", info.f.cmd);
+	}
+	close(pfd[1]);
+	if (pid < 0) {
+		close(pfd[0]);
+	} else {
 		fcntl(pfd[0], F_SETFL, O_NONBLOCK);
 		info.fd = pfd[0];
 		info.i = info.lastsep = 0;
 		info.open = true;
-	} else if (pid == 0) {
-		close(pfd[0]);
-		dup2(pfd[1], 1);
-		execl(info.cmd, info.cmd, files[fileidx].name, NULL);
-		warn("could not exec: %s", info.cmd);
-		exit(EXIT_FAILURE);
 	}
 }
 
@@ -319,7 +294,8 @@ void load_image(int new)
 	if (new < 0 || new >= filecnt)
 		return;
 
-	win_set_cursor(&win, CURSOR_WATCH);
+	if (win.xwin != None)
+		win_set_cursor(&win, CURSOR_WATCH);
 	reset_timeout(slideshow);
 
 	if (new != current)
@@ -333,7 +309,7 @@ void load_image(int new)
 		else if (new > 0 && new < fileidx)
 			new--;
 	}
-	files[new].warn = false;
+	files[new].flags &= ~FF_WARN;
 	fileidx = current = new;
 
 	info.open = false;
@@ -359,7 +335,7 @@ void bar_put(win_bar_t *bar, const char *fmt, ...)
 void update_info(void)
 {
 	unsigned int i, fn, fw;
-	char title[TITLE_LEN];
+	char title[256];
 	const char * mark;
 	bool ow_info;
 	win_bar_t *l = &win.bar.l, *r = &win.bar.r;
@@ -376,12 +352,15 @@ void update_info(void)
 	if (win.bar.h == 0)
 		return;
 	for (fw = 0, i = filecnt; i > 0; fw++, i /= 10);
-	mark = files[fileidx].marked ? "* " : "";
+	mark = files[fileidx].flags & FF_MARK ? "* " : "";
 	l->p = l->buf;
 	r->p = r->buf;
 	if (mode == MODE_THUMB) {
 		if (tns.loadnext < tns.end) {
-			bar_put(l, "Loading... %0*d", fw, MAX(tns.loadnext, 1));
+			bar_put(l, "Loading... %0*d", fw, tns.loadnext + 1);
+			ow_info = false;
+		} else if (tns.initnext < filecnt) {
+			bar_put(l, "Caching... %0*d", fw, tns.initnext + 1);
 			ow_info = false;
 		} else {
 			ow_info = true;
@@ -399,7 +378,7 @@ void update_info(void)
 			bar_put(r, "%0*d/%d | ", fn, img.multi.sel + 1, img.multi.cnt);
 		}
 		bar_put(r, "%0*d/%d", fw, fileidx + 1, filecnt);
-		ow_info = info.cmd == NULL;
+		ow_info = info.f.err != 0;
 	}
 	if (ow_info) {
 		fn = strlen(files[fileidx].name);
@@ -449,7 +428,7 @@ void reset_cursor(void)
 			}
 		}
 	} else {
-		if (tns.loadnext < tns.end)
+		if (tns.loadnext < tns.end || tns.initnext < filecnt)
 			cursor = CURSOR_WATCH;
 		else
 			cursor = CURSOR_ARROW;
@@ -479,16 +458,17 @@ void clear_resize(void)
 void run_key_handler(const char *key, unsigned int mask)
 {
 	pid_t pid;
-	int i, j, retval, status;
-	int fcnt = mode == MODE_THUMB && markcnt > 0 ? markcnt : 1;
+	FILE *pfs;
+	bool marked = mode == MODE_THUMB && markcnt > 0;
 	bool changed = false;
-	char **args, kstr[32], oldbar[BAR_L_LEN];
-	struct stat *oldst, newst;
-	struct { int fn; struct stat st; } *finfo;
+	int f, i, pfd[2], status;
+	int fcnt = marked ? markcnt : 1;
+	char kstr[32];
+	struct stat *oldst, st;
 
-	if (keyhandler.cmd == NULL) {
+	if (keyhandler.f.err != 0) {
 		if (!keyhandler.warned) {
-			warn("key handler not installed");
+			error(0, keyhandler.f.err, "%s", keyhandler.f.cmd);
 			keyhandler.warned = true;
 		}
 		return;
@@ -496,55 +476,63 @@ void run_key_handler(const char *key, unsigned int mask)
 	if (key == NULL)
 		return;
 
-	finfo = s_malloc(fcnt * sizeof(*finfo));
-	args = s_malloc((fcnt + 3) * sizeof(*args));
-	args[0] = keyhandler.cmd;
-	args[1] = kstr;
-	args[fcnt+2] = NULL;
-	if (mode == MODE_IMAGE || markcnt == 0) {
-		finfo[0].fn = fileidx;
-		stat(files[fileidx].path, &finfo[0].st);
-		args[2] = (char*) files[fileidx].path;
-	} else for (i = j = 0; i < filecnt; i++) {
-		if (files[i].marked) {
-			finfo[j].fn = i;
-			stat(files[i].path, &finfo[j++].st);
-			args[j+1] = (char*) files[i].path;
-		}
+	if (pipe(pfd) < 0) {
+		error(0, errno, "pipe");
+		return;
 	}
+	if ((pfs = fdopen(pfd[1], "w")) == NULL) {
+		error(0, errno, "open pipe");
+		close(pfd[0]), close(pfd[1]);
+		return;
+	}
+	oldst = emalloc(fcnt * sizeof(*oldst));
+
+	strncpy(win.bar.l.buf, "Running key handler...", win.bar.l.size);
+	win_draw(&win);
+	win_set_cursor(&win, CURSOR_WATCH);
+
 	snprintf(kstr, sizeof(kstr), "%s%s%s%s",
 	         mask & ControlMask ? "C-" : "",
 	         mask & Mod1Mask    ? "M-" : "",
 	         mask & ShiftMask   ? "S-" : "", key);
 
-	memcpy(oldbar, win.bar.l.buf, sizeof(oldbar));
-	strncpy(win.bar.l.buf, "Running key handler...", win.bar.l.size);
-	win_draw(&win);
-	win_set_cursor(&win, CURSOR_WATCH);
-
 	if ((pid = fork()) == 0) {
-		execv(keyhandler.cmd, args);
-		warn("could not exec key handler");
-		exit(EXIT_FAILURE);
-	} else if (pid < 0) {
-		warn("could not fork key handler");
+		close(pfd[1]);
+		dup2(pfd[0], 0);
+		execl(keyhandler.f.cmd, keyhandler.f.cmd, kstr, NULL);
+		error(EXIT_FAILURE, errno, "exec: %s", keyhandler.f.cmd);
+	}
+	close(pfd[0]);
+	if (pid < 0) {
+		error(0, errno, "fork");
+		fclose(pfs);
 		goto end;
 	}
-	waitpid(pid, &status, 0);
-	retval = WEXITSTATUS(status);
-	if (WIFEXITED(status) == 0 || retval != 0)
-		warn("key handler exited with non-zero return value: %d", retval);
 
-	for (i = 0; i < fcnt; i++) {
-		oldst = &finfo[i].st;
-		if (stat(files[finfo[i].fn].path, &newst) != 0 ||
-				memcmp(&oldst->st_mtime, &newst.st_mtime, sizeof(newst.st_mtime)) != 0)
-		{
-			if (tns.thumbs != NULL) {
-				tns_unload(&tns, finfo[i].fn);
-				tns.loadnext = MIN(tns.loadnext, finfo[i].fn);
+	for (f = i = 0; f < fcnt; i++) {
+		if ((marked && (files[i].flags & FF_MARK)) || (!marked && i == fileidx)) {
+			stat(files[i].path, &oldst[f]);
+			fprintf(pfs, "%s\n", files[i].path);
+			f++;
+		}
+	}
+	fclose(pfs);
+	waitpid(pid, &status, 0);
+	if (!WIFEXITED(status) || WEXITSTATUS(status) != 0)
+		error(0, 0, "%s: Exited abnormally", keyhandler.f.cmd);
+
+	for (f = i = 0; f < fcnt; i++) {
+		if ((marked && (files[i].flags & FF_MARK)) || (!marked && i == fileidx)) {
+			if (stat(files[i].path, &st) != 0 ||
+				  memcmp(&oldst[f].st_mtime, &st.st_mtime, sizeof(st.st_mtime)) != 0)
+			{
+				if (tns.thumbs != NULL) {
+					tns_unload(&tns, i);
+					tns.loadnext = MIN(tns.loadnext, i);
+				}
+				changed = true;
 			}
-			changed = true;
+			f++;
 		}
 	}
 end:
@@ -552,14 +540,14 @@ end:
 		if (changed) {
 			img_close(&img, true);
 			load_image(fileidx);
-		} else if (info.cmd != NULL) {
-			memcpy(win.bar.l.buf, oldbar, win.bar.l.size);
+		} else if (info.f.err == 0) {
+			info.open = false;
+			open_info();
 		}
 	}
+	free(oldst);
 	reset_cursor();
 	redraw();
-	free(finfo);
-	free(args);
 }
 
 #define MODMASK(mask) ((mask) & (ShiftMask|ControlMask|Mod1Mask))
@@ -572,9 +560,6 @@ void on_keypress(XKeyEvent *kev)
 	char key;
 	bool dirty = false;
 
-	if (kev == NULL)
-		return;
-
 	if (kev->state & ShiftMask) {
 		kev->state &= ~ShiftMask;
 		XLookupString(kev, &key, 1, &shksym, NULL);
@@ -615,9 +600,6 @@ void on_buttonpress(XButtonEvent *bev)
 	bool dirty = false;
 	static Time firstclick;
 
-	if (bev == NULL)
-		return;
-
 	if (mode == MODE_IMAGE) {
 		win_set_cursor(&win, CURSOR_ARROW);
 		set_timeout(reset_cursor, TO_CURSOR_HIDE, true);
@@ -657,9 +639,9 @@ void on_buttonpress(XButtonEvent *bev)
 				break;
 			case Button3:
 				if ((sel = tns_translate(&tns, bev->x, bev->y)) >= 0) {
-					files[sel].marked = !files[sel].marked;
-					markcnt += files[sel].marked ? 1 : -1;
-					tns_mark(&tns, sel, files[sel].marked);
+					files[sel].flags ^= FF_MARK;
+					markcnt += files[sel].flags & FF_MARK ? 1 : -1;
+					tns_mark(&tns, sel, !!(files[sel].flags & FF_MARK));
 					redraw();
 				}
 				break;
@@ -679,26 +661,29 @@ void run(void)
 	int xfd;
 	fd_set fds;
 	struct timeval timeout;
-	bool discard, load_thumb, to_set;
+	bool discard, init_thumb, load_thumb, to_set;
 	XEvent ev, nextev;
 
 	while (true) {
 		to_set = check_timeouts(&timeout);
+		init_thumb = mode == MODE_THUMB && tns.initnext < filecnt;
 		load_thumb = mode == MODE_THUMB && tns.loadnext < tns.end;
 
-		if ((load_thumb || to_set || info.fd != -1) &&
+		if ((init_thumb || load_thumb || to_set || info.fd != -1) &&
 		    XPending(win.env.dpy) == 0)
 		{
 			if (load_thumb) {
 				set_timeout(redraw, TO_REDRAW_THUMBS, false);
-				if (!tns_load(&tns, tns.loadnext, false)) {
+				if (!tns_load(&tns, tns.loadnext, false, false)) {
 					remove_file(tns.loadnext, false);
 					tns.dirty = true;
 				}
-				while (tns.loadnext < tns.end && tns.thumbs[tns.loadnext].im != NULL)
-					tns.loadnext++;
 				if (tns.loadnext >= tns.end)
 					redraw();
+			} else if (init_thumb) {
+				set_timeout(redraw, TO_REDRAW_THUMBS, false);
+				if (!tns_load(&tns, tns.initnext, false, true))
+					remove_file(tns.initnext, false);
 			} else {
 				xfd = ConnectionNumber(win.env.dpy);
 				FD_ZERO(&fds);
@@ -785,6 +770,8 @@ int main(int argc, char **argv)
 	struct stat fstats;
 	r_dir_t dir;
 
+	signal(SIGPIPE, SIG_IGN);
+
 	parse_options(argc, argv);
 
 	if (options->clean_cache) {
@@ -799,16 +786,18 @@ int main(int argc, char **argv)
 	}
 
 	if (options->recursive || options->from_stdin)
-		filecnt = FILENAME_CNT;
+		filecnt = 1024;
 	else
 		filecnt = options->filecnt;
 
-	files = (fileinfo_t*) s_malloc(filecnt * sizeof(fileinfo_t));
+	files = emalloc(filecnt * sizeof(*files));
+	memset(files, 0, filecnt * sizeof(*files));
 	fileidx = 0;
 
 	if (options->from_stdin) {
+		n = 0;
 		filename = NULL;
-		while ((len = get_line(&filename, &n, stdin)) > 0) {
+		while ((len = getline(&filename, &n, stdin)) > 0) {
 			if (filename[len-1] == '\n')
 				filename[len-1] = '\0';
 			check_add_file(filename, true);
@@ -820,18 +809,18 @@ int main(int argc, char **argv)
 		filename = options->filenames[i];
 
 		if (stat(filename, &fstats) < 0) {
-			warn("could not stat file: %s", filename);
+			error(0, errno, "%s", filename);
 			continue;
 		}
 		if (!S_ISDIR(fstats.st_mode)) {
 			check_add_file(filename, true);
 		} else {
 			if (!options->recursive) {
-				warn("ignoring directory: %s", filename);
+				error(0, 0, "%s: Is a directory", filename);
 				continue;
 			}
 			if (r_opendir(&dir, filename) < 0) {
-				warn("could not open directory: %s", filename);
+				error(0, errno, "%s", filename);
 				continue;
 			}
 			start = fileidx;
@@ -845,10 +834,8 @@ int main(int argc, char **argv)
 		}
 	}
 
-	if (fileidx == 0) {
-		fprintf(stderr, "sxiv: no valid image file given, aborting\n");
-		exit(EXIT_FAILURE);
-	}
+	if (fileidx == 0)
+		error(EXIT_FAILURE, 0, "No valid image file given, aborting");
 
 	filecnt = fileidx;
 	fileidx = options->startnum < filecnt ? options->startnum : 0;
@@ -861,28 +848,26 @@ int main(int argc, char **argv)
 		dsuffix = "/.config";
 	}
 	if (homedir != NULL) {
-		char **cmd[] = { &info.cmd, &keyhandler.cmd };
+		extcmd_t *cmd[] = { &info.f, &keyhandler.f };
 		const char *name[] = { "image-info", "key-handler" };
 
 		for (i = 0; i < ARRLEN(cmd); i++) {
-			len = strlen(homedir) + strlen(dsuffix) + strlen(name[i]) + 12;
-			*cmd[i] = (char*) s_malloc(len);
-			snprintf(*cmd[i], len, "%s%s/sxiv/exec/%s", homedir, dsuffix, name[i]);
-			if (access(*cmd[i], X_OK) != 0) {
-				free(*cmd[i]);
-				*cmd[i] = NULL;
-			}
+			n = strlen(homedir) + strlen(dsuffix) + strlen(name[i]) + 12;
+			cmd[i]->cmd = (char*) emalloc(n);
+			snprintf(cmd[i]->cmd, n, "%s%s/sxiv/exec/%s", homedir, dsuffix, name[i]);
+			if (access(cmd[i]->cmd, X_OK) != 0)
+				cmd[i]->err = errno;
 		}
 	} else {
-		warn("could not locate exec directory");
+		error(0, 0, "Exec directory not found");
 	}
 	info.fd = -1;
 
 	if (options->thumb_mode) {
 		mode = MODE_THUMB;
 		tns_init(&tns, files, &filecnt, &fileidx, &win);
-		while (!tns_load(&tns, 0, false))
-			remove_file(0, false);
+		while (!tns_load(&tns, fileidx, false, false))
+			remove_file(fileidx, false);
 	} else {
 		mode = MODE_IMAGE;
 		tns.thumbs = NULL;
@@ -891,10 +876,11 @@ int main(int argc, char **argv)
 	win_open(&win);
 	win_set_cursor(&win, CURSOR_WATCH);
 
+	atexit(cleanup);
+
 	set_timeout(redraw, 25, false);
 
 	run();
-	cleanup();
 
 	return 0;
 }
