@@ -26,7 +26,6 @@
 #include "window.h"
 
 typedef struct {
-	const fileinfo_t *file;
 	Imlib_Image im;
 	int w;
 	int h;
@@ -35,27 +34,34 @@ typedef struct {
 } thumb_t;
 
 typedef struct {
+	fileinfo_t *files;
 	thumb_t *thumbs;
-	int cap;
-	int cnt;
-	int first;
-	int sel;
+	const int *cnt;
+	int *sel;
+	int initnext;
+	int loadnext;
+	int first, end;
+	int r_first, r_end;
 
 	win_t *win;
 	int x;
 	int y;
 	int cols;
 	int rows;
+	int zl;
+	int bw;
+	int dim;
 
 	bool dirty;
 } tns_t;
 
 void tns_clean_cache(tns_t*);
 
-void tns_init(tns_t*, int, win_t*);
-void tns_free(tns_t*);
+void tns_init(tns_t*, fileinfo_t*, const int*, int*, win_t*);
+CLEANUP void tns_free(tns_t*);
 
-bool tns_load(tns_t*, int, const fileinfo_t*, bool, bool);
+bool tns_load(tns_t*, int, bool, bool);
+void tns_unload(tns_t*, int);
 
 void tns_render(tns_t*);
 void tns_mark(tns_t*, int, bool);
@@ -64,6 +70,8 @@ void tns_highlight(tns_t*, int, bool);
 bool tns_move_selection(tns_t*, direction_t, int);
 bool tns_scroll(tns_t*, direction_t, bool);
 
+bool tns_zoom(tns_t*, int);
+
 int tns_translate(tns_t*, int, int);
 
 #endif /* THUMBS_H */
