@@ -21,6 +21,8 @@
 
 #include <stdbool.h>
 
+#define CLEANUP
+
 typedef enum {
 	BO_BIG_ENDIAN,
 	BO_LITTLE_ENDIAN
@@ -32,10 +34,10 @@ typedef enum {
 } appmode_t;
 
 typedef enum {
-	DIR_LEFT,
-	DIR_RIGHT,
-	DIR_UP,
-	DIR_DOWN
+	DIR_LEFT  = 1,
+	DIR_RIGHT = 2,
+	DIR_UP    = 4,
+	DIR_DOWN  = 8
 } direction_t;
 
 typedef enum {
@@ -64,12 +66,17 @@ typedef enum {
 	CURSOR_WATCH
 } cursor_t;
 
+typedef enum {
+	FF_WARN    = 1,
+	FF_MARK    = 2,
+	FF_TN_INIT = 4
+} fileflags_t;
+
 typedef struct {
 	const char *name; /* as given by user */
 	const char *path; /* always absolute */
 	const char *base;
-	bool loaded;
-	bool marked;
+	fileflags_t flags;
 } fileinfo_t;
 
 /* timeouts in milliseconds: */
